from collections.abc import Callable
from typing import Annotated, Any, Optional, cast

import numpy as np
from beartype import beartype
from beartype.vale import Is, IsAttr, IsEqual
from numpy import (
    bool_,
    dtype,
    float64,
    int64,
    isfinite,
    isnan,
    log,
    object_,
    rint,
    unravel_index,
)
from numpy.random import default_rng
from numpy.typing import NDArray

from utilities.beartype import NDim0, NDim1, NDim2, NDim3

# dtypes
datetime64D = dtype("datetime64[D]")  # noqa: N816
datetime64Y = dtype("datetime64[Y]")  # noqa: N816
datetime64ns = dtype("datetime64[ns]")

# dtype checkers
DTypeB = IsAttr["dtype", IsEqual[bool]]
DTypeDD = IsAttr["dtype", IsEqual[datetime64D]]
DTypeDY = IsAttr["dtype", IsEqual[datetime64Y]]
DTypeDns = IsAttr["dtype", IsEqual[datetime64ns]]
DTypeF = IsAttr["dtype", IsEqual[float]]
DTypeI = IsAttr["dtype", IsEqual[int]]
DTypeO = IsAttr["dtype", IsEqual[object]]

# annotated; dtype
NDArrayB = NDArray[bool_]
NDArrayDD = NDArray[cast(Any, datetime64D)]
NDArrayDY = NDArray[cast(Any, datetime64Y)]
NDArrayDns = NDArray[cast(Any, datetime64ns)]
NDArrayF = NDArray[float64]
NDArrayI = NDArray[int64]
NDArrayO = NDArray[object_]

# annotated; ndim
NDArray0 = Annotated[NDArray[Any], NDim0]
NDArray1 = Annotated[NDArray[Any], NDim1]
NDArray2 = Annotated[NDArray[Any], NDim2]
NDArray3 = Annotated[NDArray[Any], NDim3]

# annotated; dtype & ndim
NDArrayB0 = Annotated[NDArrayB, NDim0]
NDArrayDD0 = Annotated[NDArrayDD, NDim0]
NDArrayDY0 = Annotated[NDArrayDY, NDim0]
NDArrayDns0 = Annotated[NDArrayDns, NDim0]
NDArrayF0 = Annotated[NDArrayF, NDim0]
NDArrayI0 = Annotated[NDArrayI, NDim0]
NDArrayO0 = Annotated[NDArrayO, NDim0]

NDArrayB1 = Annotated[NDArrayB, NDim1]
NDArrayDD1 = Annotated[NDArrayDD, NDim1]
NDArrayDY1 = Annotated[NDArrayDY, NDim1]
NDArrayDns1 = Annotated[NDArrayDns, NDim1]
NDArrayF1 = Annotated[NDArrayF, NDim1]
NDArrayI1 = Annotated[NDArrayI, NDim1]
NDArrayO1 = Annotated[NDArrayO, NDim1]

NDArrayB2 = Annotated[NDArrayB, NDim2]
NDArrayDD2 = Annotated[NDArrayDD, NDim2]
NDArrayDY2 = Annotated[NDArrayDY, NDim2]
NDArrayDns2 = Annotated[NDArrayDns, NDim2]
NDArrayF2 = Annotated[NDArrayF, NDim2]
NDArrayI2 = Annotated[NDArrayI, NDim2]
NDArrayO2 = Annotated[NDArrayO, NDim2]

NDArrayB3 = Annotated[NDArrayB, NDim3]
NDArrayDD3 = Annotated[NDArrayDD, NDim3]
NDArrayDY3 = Annotated[NDArrayDY, NDim3]
NDArrayDns3 = Annotated[NDArrayDns, NDim3]
NDArrayF3 = Annotated[NDArrayF, NDim3]
NDArrayI3 = Annotated[NDArrayI, NDim3]
NDArrayO3 = Annotated[NDArrayO, NDim3]


# checks


@beartype
def is_at_least(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x >= y."""
    return (x >= y) | _is_close(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan)


@beartype
def is_at_least_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x >= y or x == nan."""
    return is_at_least(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


@beartype
def is_at_most(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x <= y."""
    return (x <= y) | _is_close(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan)


@beartype
def is_at_most_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x <= y or x == nan."""
    return is_at_most(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


@beartype
def is_between(
    x: Any,
    low: Any,
    high: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
    low_equal_nan: bool = False,
    high_equal_nan: bool = False,
) -> Any:
    """Check if low <= x <= high."""
    return is_at_least(
        x, low, rtol=rtol, atol=atol, equal_nan=equal_nan or low_equal_nan
    ) & is_at_most(x, high, rtol=rtol, atol=atol, equal_nan=equal_nan or high_equal_nan)


@beartype
def is_between_or_nan(
    x: Any,
    low: Any,
    high: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
    low_equal_nan: bool = False,
    high_equal_nan: bool = False,
) -> Any:
    """Check if low <= x <= high or x == nan."""
    return is_between(
        x,
        low,
        high,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan,
        low_equal_nan=low_equal_nan,
        high_equal_nan=high_equal_nan,
    ) | isnan(x)


@beartype
def is_finite_and_integral(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x < inf and x == int(x)."""
    return isfinite(x) and is_integral(x, rtol=rtol, atol=atol)


@beartype
def is_finite_and_integral_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x < inf and x == int(x), or x == nan."""
    return is_finite_and_integral(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_finite_and_negative(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x < 0."""
    return isfinite(x) & is_negative(x, rtol=rtol, atol=atol)


@beartype
def is_finite_and_negative_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x < 0 or x == nan."""
    return is_finite_and_negative(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_finite_and_non_negative(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if 0 <= x < inf."""
    return isfinite(x) & is_non_negative(x, rtol=rtol, atol=atol)


@beartype
def is_finite_and_non_negative_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if 0 <= x < inf or x == nan."""
    return is_finite_and_non_negative(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_finite_and_non_positive(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x <= 0."""
    return isfinite(x) & is_non_positive(x, rtol=rtol, atol=atol)


@beartype
def is_finite_and_non_positive_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x <= 0 or x == nan."""
    return is_finite_and_non_positive(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_finite_and_non_zero(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if -inf < x < inf, x != 0."""
    return isfinite(x) & is_non_zero(x, rtol=rtol, atol=atol)


@beartype
def is_finite_and_non_zero_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x != 0 or x == nan."""
    return is_finite_and_non_zero(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_finite_and_positive(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if 0 < x < inf."""
    return isfinite(x) & is_positive(x, rtol=rtol, atol=atol)


@beartype
def is_finite_and_positive_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if 0 < x < inf or x == nan."""
    return is_finite_and_positive(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_greater_than(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x > y."""
    return ((x > y) & ~_is_close(x, y, rtol=rtol, atol=atol)) | (
        equal_nan & isnan(x) & isnan(y)
    )


@beartype
def is_greater_than_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x > y or x == nan."""
    return is_greater_than(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


@beartype
def is_integral(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x == int(x)."""
    return _is_close(x, rint(x), rtol=rtol, atol=atol)


@beartype
def is_integral_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x == int(x) or x == nan."""
    return is_integral(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_less_than(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x < y."""
    return ((x < y) & ~_is_close(x, y, rtol=rtol, atol=atol)) | (
        equal_nan & isnan(x) & isnan(y)
    )


@beartype
def is_less_than_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x < y or x == nan."""
    return is_less_than(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


@beartype
def is_negative(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x < 0."""
    return is_less_than(x, 0.0, rtol=rtol, atol=atol)


@beartype
def is_negative_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x < 0 or x == nan."""
    return is_negative(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_non_negative(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x >= 0."""
    return is_at_least(x, 0.0, rtol=rtol, atol=atol)


@beartype
def is_non_negative_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x >= 0 or x == nan."""
    return is_non_negative(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_non_positive(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x <= 0."""
    return is_at_most(x, 0.0, rtol=rtol, atol=atol)


@beartype
def is_non_positive_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x <=0 or x == nan."""
    return is_non_positive(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_non_zero(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x != 0."""
    return ~_is_close(x, 0.0, rtol=rtol, atol=atol)


@beartype
def is_non_zero_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x != 0 or x == nan."""
    return is_non_zero(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_positive(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x > 0."""
    return is_greater_than(x, 0, rtol=rtol, atol=atol)


@beartype
def is_positive_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x > 0 or x == nan."""
    return is_positive(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def is_zero(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x == 0."""
    return _is_close(x, 0.0, rtol=rtol, atol=atol)


@beartype
def is_zero_or_nan(
    x: Any, /, *, rtol: Optional[float] = None, atol: Optional[float] = None
) -> Any:
    """Check if x > 0 or x == nan."""
    return is_zero(x, rtol=rtol, atol=atol) | isnan(x)


@beartype
def _is_close(
    x: Any,
    y: Any,
    /,
    *,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    equal_nan: bool = False,
) -> Any:
    """Check if `x` is close to `y`."""
    return np.isclose(
        x,
        y,
        **({} if rtol is None else {"rtol": rtol}),
        **({} if atol is None else {"atol": atol}),
        equal_nan=equal_nan,
    )


# annotated; float & checks


@beartype
def _lift(check: Callable[..., Any], /) -> Any:
    """Lift a check to work on a subset of a float array."""
    rng = default_rng()

    @beartype
    def predicate(array: NDArrayF, /) -> bool:
        if (size := array.size) == 0:
            return True
        if size == 1:
            return check(array).item()
        num_samples = round(log(size))
        indices = rng.integers(0, size, size=num_samples)
        sample = array[unravel_index(indices, array.shape)]
        return check(sample).all().item()

    return Is[cast(Any, predicate)]


NDArrayFFinInt = Annotated[NDArrayF, _lift(is_finite_and_integral)]
NDArrayFFinIntNan = Annotated[NDArrayF, _lift(is_finite_and_integral_or_nan)]
NDArrayFFinNeg = Annotated[NDArrayF, _lift(is_finite_and_negative)]
NDArrayFFinNegNan = Annotated[NDArrayF, _lift(is_finite_and_negative_or_nan)]
NDArrayFFinNonNeg = Annotated[NDArrayF, _lift(is_finite_and_non_negative)]
NDArrayFFinNonNegNan = Annotated[NDArrayF, _lift(is_finite_and_non_negative_or_nan)]
NDArrayFFinNonPos = Annotated[NDArrayF, _lift(is_finite_and_non_positive)]
NDArrayFFinNonPosNan = Annotated[NDArrayF, _lift(is_finite_and_non_positive_or_nan)]
NDArrayFFinNonZr = Annotated[NDArrayF, _lift(is_finite_and_non_zero)]
NDArrayFFinNonZrNan = Annotated[NDArrayF, _lift(is_finite_and_non_zero_or_nan)]
NDArrayFFinPos = Annotated[NDArrayF, _lift(is_finite_and_positive)]
NDArrayFFinPosNan = Annotated[NDArrayF, _lift(is_finite_and_positive_or_nan)]
NDArrayFInt = Annotated[NDArrayF, _lift(is_integral)]
NDArrayFIntNan = Annotated[NDArrayF, _lift(is_integral_or_nan)]
NDArrayFNeg = Annotated[NDArrayF, _lift(is_negative)]
NDArrayFNegNan = Annotated[NDArrayF, _lift(is_negative_or_nan)]
NDArrayFNonNeg = Annotated[NDArrayF, _lift(is_non_negative)]
NDArrayFNonNegNan = Annotated[NDArrayF, _lift(is_non_negative_or_nan)]
NDArrayFNonPos = Annotated[NDArrayF, _lift(is_non_positive)]
NDArrayFNonPosNan = Annotated[NDArrayF, _lift(is_non_positive_or_nan)]
NDArrayFNonZr = Annotated[NDArrayF, _lift(is_non_zero)]
NDArrayFNonZrNan = Annotated[NDArrayF, _lift(is_non_zero_or_nan)]
NDArrayFPos = Annotated[NDArrayF, _lift(is_positive)]
NDArrayFPosNan = Annotated[NDArrayF, _lift(is_positive_or_nan)]
NDArrayFZr = Annotated[NDArrayF, _lift(is_zero)]
NDArrayFZrNan = Annotated[NDArrayF, _lift(is_zero_or_nan)]
